/*
-----------------------------------------------------------------------------
This source file is part of OGRE
    (Object-oriented Graphics Rendering Engine)
For the latest info, see http://www.ogre3d.org/

Copyright (c) 2000-2005 The OGRE Team
Also see acknowledgements in Readme.html

You may use this sample code for anything you like, it is not covered by the
LGPL like the rest of the engine.
-----------------------------------------------------------------------------
*/

/**
    \file 
        Lighting.h
    \brief
        Shows lighting support in OGRE. Also demonstrates the use of billboards
        and controllers for automatic time-relative behaviour.
*/


#include "ExampleApplication.h"
#include "OgreIlluminationManager.h"
#include "SpriteParticleRenderer.h"

bool HighQuality = false;
float causize = 4;
float cauintens = 25;

// Listener class for frame updates
class RaytraceDemoListener : public ExampleFrameListener
{
protected:
	SceneNode* goldSphereNode;
	SceneNode* silverSphereNode;
	SceneNode* copperSphereNode;
	SceneNode* aluSphereNode;
	SceneNode* lightNode;
	unsigned long framecount;
	
public:
    RaytraceDemoListener(RenderWindow* window, Camera* maincam,SceneNode* lightNode,
																SceneNode* goldSphereNode,
																SceneNode* silverSphereNode,
																SceneNode* copperSphereNode,
																SceneNode* aluSphereNode)
        :ExampleFrameListener(window, maincam) 
				 
    {
		framecount = 0;
		this->goldSphereNode = goldSphereNode;
		this->silverSphereNode = silverSphereNode;
		this->copperSphereNode = copperSphereNode;
		this->aluSphereNode = aluSphereNode;
		this->lightNode = lightNode;
    }
	bool processUnbufferedKeyInput(const FrameEvent& evt)
    {
		MaterialPtr mat = MaterialManager::getSingleton().getByName("GameTools/Cau");
		GpuProgramParameters* Vparams = mat->getTechnique(0)->getPass(0)->getVertexProgramParameters().getPointer();
		GpuProgramParameters* Fparams = mat->getTechnique(0)->getPass(0)->getFragmentProgramParameters().getPointer();
		
		if (mInputDevice->isKeyDown(KC_SUBTRACT))
        {
			cauintens -= 1;
			cauintens = std::max(cauintens, 1.0f);
			float intensity = pow( cauintens / 100.0f, 2 );
			Fparams->setNamedConstant("CauSpriteIntens", intensity);			
		}
		if (mInputDevice->isKeyDown(KC_ADD))
        {
			cauintens += 1;
			cauintens = std::min(cauintens, 100.0f);
			float intensity = pow( cauintens / 100.0f, 2 );
			Fparams->setNamedConstant("CauSpriteIntens", intensity);				
		}
		if (mInputDevice->isKeyDown(KC_MULTIPLY))
        {
			causize = std::max(causize - 0.2, 0.0);
			Vparams->setNamedConstant("CauSpriteSize", causize);			
		}
		if (mInputDevice->isKeyDown(KC_DIVIDE))
        {
			causize = causize + 0.2;
			Vparams->setNamedConstant("CauSpriteSize", causize);			
		}

		return ExampleFrameListener::processUnbufferedKeyInput(evt);
	}
    bool frameStarted(const FrameEvent& evt)
    {
		static bool first = true;
		if(first)
		{
			MaterialPtr mat = MaterialManager::getSingleton().getByName("GameTools/Cau");
			GpuProgramParameters* Vparams = mat->getTechnique(0)->getPass(0)->getVertexProgramParameters().getPointer();
			GpuProgramParameters* Fparams = mat->getTechnique(0)->getPass(0)->getFragmentProgramParameters().getPointer();
			float intensity = pow( cauintens / 100.0f, 2 );
			Fparams->setNamedConstant("CauSpriteIntens", intensity);			
			Vparams->setNamedConstant("CauSpriteSize", causize);
			first = false;
		}

		lightNode->setPosition( 100.0 * Vector3(cos(-(float)framecount/500.0),2.0,sin(-(float)framecount/500.0)));
		//lightNode->setDirection(-lightNode->getPosition());
		
		goldSphereNode->setPosition(   100.0 * 
								Vector3(
										cos((float)framecount/700.0),
										1.0,
										sin((float)framecount/700.0)));
		silverSphereNode->setPosition( 100.0 * 
								Vector3(
										cos((float)framecount/700.0 + Math::PI / 2.0),
										1.0,
										sin((float)framecount/700.0 + Math::PI / 2.0)));
		copperSphereNode->setPosition( 100.0 * 
								Vector3(
										cos((float)framecount/700.0 + Math::PI ),
										1.0,
										sin((float)framecount/700.0 + Math::PI)));
		aluSphereNode->setPosition(    100.0 * 
								Vector3(
										cos((float)framecount/700.0 + 3 * Math::PI / 2.0), 
										1.0,
										sin((float)framecount/700.0+ 3 * Math::PI / 2.0)));
		
		bool result = ExampleFrameListener::frameStarted(evt);
       framecount++;
        return result;     
    }

	

};

/** Application class */
class RaytraceDemoApplication : public ExampleApplication
{

public:
	virtual void go(void)
    {
        if (!setup())
            return;

        renderScene();
		
        // clean up
        destroyScene();
    }

protected:
	SceneNode* kupolaNode;
	SceneNode* mainLightNode;
	SceneNode* goldSphereNode;
	SceneNode* silverSphereNode;
	SceneNode* copperSphereNode;
	SceneNode* aluSphereNode;

		
	void renderScene(){
		
		mWindow->resetStatistics();
		while(true){		

			
			#if OGRE_PLATFORM == OGRE_PLATFORM_WIN32            
			// Pump events on Win32
    		MSG  msg;
			while( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
			{
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
			
			#endif
			
			if(!mRoot->_fireFrameStarted()){
				break;
			}
			unsigned long framenum = mRoot->getCurrentFrameNumber();
			
			OgreIlluminationManager::getSingleton().update(framenum, mWindow);

			//Root::getSingleton().renderOneFrame();
			
			
			
			
			mRoot->_fireFrameEnded();	
			
		}
	}

	void createPostproc()
	{
		CompositorManager::getSingleton().addCompositor(mWindow->getViewport(0),
														"GameTools/Glow");
		CompositorManager::getSingleton().setCompositorEnabled(mWindow->getViewport(0),
														"GameTools/Glow", true);
/*
		CompositorManager::getSingleton().addCompositor(mWindow->getViewport(0),
														"GameTools/ToneMap");
		CompositorManager::getSingleton().setCompositorEnabled(mWindow->getViewport(0),
														"GameTools/ToneMap", true);*/
	}
    

	void createScene(void)
    {

		createPostproc();
		
		OgreIlluminationManager::getSingleton().setMainCamera(mCamera);
		OgreIlluminationManager::getSingleton().setMainViewport(mWindow->getViewport(0));
		
		if(! HighQuality)
			OgreIlluminationManager::getSingleton().setMaxJoinRadius(420);
		else
			OgreIlluminationManager::getSingleton().setMaxJoinRadius(0.1);


		Root::getSingleton()._setCurrentSceneManager(mSceneMgr);
		mCamera->setPosition(0,100,240);
		mCamera->setFOVy(Radian(Degree(80)));
		//mCamera->setFarClipDistance(1000);
		// Set ambient light
        mSceneMgr->setAmbientLight(ColourValue(1, 1, 1));

		// Skybox
        mSceneMgr->setSkyBox(true, "Examples/MorningSkyBox");

		SceneNode* rootNode = mSceneMgr->getRootSceneNode();

      
		//init objects
		Entity* kupola = mSceneMgr->createEntity("kupola", "kupola.mesh");
		Entity* athene = mSceneMgr->createEntity("athene", "head.mesh");
		Entity* goldSphere = mSceneMgr->createEntity("goldSphere", "head.mesh");
		Entity* silverSphere = mSceneMgr->createEntity("silverSphere", "head.mesh");
		Entity* copperSphere = mSceneMgr->createEntity("copperSphere", "head.mesh");
		Entity* aluSphere = mSceneMgr->createEntity("aluSphere", "head.mesh");

		//lod
		//MeshPtr amesh = athene->getMesh();
		//Mesh::LodDistanceList distances;
		//distances.push_back(1);
		//distances.push_back(2);
		////distances.push_back(3);
		////distances.push_back(4);
		////distances.push_back(5);
		//amesh->generateLodLevels(distances,ProgressiveMesh::VertexReductionQuota::VRQ_PROPORTIONAL,0.6);
		//athene->setMeshLodBias(50000.0,0);
		//tangents
		//amesh->buildTangentVectors(0,1);
		MeshPtr kmesh = kupola->getMesh();
		kmesh->buildTangentVectors(0,1);
		
		//init nodes
		kupolaNode = rootNode->createChildSceneNode( "kupola" );
		kupolaNode->rotate(Vector3(0,1,0),Radian(Degree(90)));
		kupolaNode->attachObject( kupola );
		kupolaNode->scale(25,21,25);		
		SceneNode* atheneNode = rootNode->createChildSceneNode("athene");
		atheneNode->attachObject(athene);
		//atheneNode->setScale(3.5,5.5,5.5);
		atheneNode->setScale(4,4,4);
		//atheneNode->setScale(0.5,0.5,0.5);
		AxisAlignedBox bbox = athene->getBoundingBox();
		atheneNode->setPosition(0, 100,0);
		

		goldSphereNode = rootNode->createChildSceneNode( "goldSphere" );
		//goldSphereNode->scale(0.3,0.3,0.3);
		goldSphereNode->scale(2,2,2);
		silverSphereNode = rootNode->createChildSceneNode( "silverSphere" );
		//silverSphereNode->scale(0.3,0.3,0.3);
		silverSphereNode->scale(2,2,2);
		copperSphereNode = rootNode->createChildSceneNode( "copperSphere" );
		//copperSphereNode->scale(0.3,0.3,0.3);
		copperSphereNode->scale(2,2,2);
		aluSphereNode = rootNode->createChildSceneNode( "aluSphere" );
		//aluSphereNode->scale(0.3,0.3,0.3);
		aluSphereNode->scale(2,2,2);

		goldSphereNode->attachObject(goldSphere);
		silverSphereNode->attachObject(silverSphere);
		aluSphereNode->attachObject(aluSphere);
		copperSphereNode->attachObject(copperSphere);

		goldSphereNode->_updateBounds();
		silverSphereNode->_updateBounds();
		aluSphereNode->_updateBounds();
		copperSphereNode->_updateBounds();

		kupolaNode->_updateBounds();
		//atheneNode->_updateBounds();

		//Init lights
		Light* mainLight = mSceneMgr->createLight("MainLight");
		mainLight->setType(Light::LT_POINT);
        mainLight->setDiffuseColour(ColourValue::White);
		mainLight->setSpecularColour(ColourValue::White);
		mainLightNode = rootNode->createChildSceneNode();
		mainLightNode->setPosition(0,0,0);		
		mainLight->setPosition(mainLightNode->getWorldPosition());
		mainLightNode->attachObject(mainLight); 

		
		//init techniques
		
		int refreshinterval;
		if(HighQuality)
			refreshinterval = 1;
		else
			refreshinterval = 0;

		bool usecalculations;
		if(HighQuality)
			usecalculations = true;
		else
			usecalculations = false;
		
		bool updateallfaces;
		if(HighQuality)
			updateallfaces = false;
		else
			updateallfaces = true;		

		{//kupola
			OgreSharedRuns* sharedruns = new OgreSharedRuns();
			OgreIlluminationManager::getSingleton().addSharedRuns(sharedruns);
			
			for(int i = 0; i < 2; i++)
			{
				OgreTechniqueGroup* group = new OgreTechniqueGroup();
				group->addSharedRun(sharedruns);

				OgreRenderable* rend = new OgreRenderable(kupola, i);			
				sharedruns->addRenderable(rend);
				sharedruns->updateBounds();		
				Material* mat = (Material*) kupola->getSubEntity(i)->getMaterial().getPointer();
			
				OgreCausticRecieverRenderTechnique* cautechn = 
				new OgreCausticRecieverRenderTechnique(1,
													"GameTools/Caustic/VertexPrograms/DefaultVS",
													"GameTools/Caustic/FragmentPrograms/DefaultPS",
													mat->getTechnique(0)->getPass(0),
													rend,
													group
													);

				OgreDepthShadowRecieverRenderTechnique* shadowtechn = 
					new OgreDepthShadowRecieverRenderTechnique(1, 
													"GameTools/ShadowMap/ShadowVS",
													"GameTools/ShadowMap/ShadowPS",
													mat->getTechnique(0)->getPass(0),
													rend,
													group
													);
				group->addRenderTechnique(cautechn);
				group->addRenderTechnique(shadowtechn);
				kupola->getSubEntity(i)->setRenderTechniqueGroup(group);
			}
		}
			

		for(int i = 0; i < 4; i++)//spheres
		{
			Entity* sphere = 0;
			if(i == 0)sphere = goldSphere;
			else if(i == 1)sphere = silverSphere;
			else if(i == 2)sphere = aluSphere;
			else if(i == 3)sphere = copperSphere;

			OgreSharedRuns* sharedruns = new OgreSharedRuns();
			OgreIlluminationManager::getSingleton().addSharedRuns(sharedruns);		
			OgreTechniqueGroup* group = new OgreTechniqueGroup();
			group->addSharedRun(sharedruns);

			OgreRenderable* rend = new OgreRenderable(sphere, 0);			
			sharedruns->addRenderable(rend);
			sharedruns->updateBounds();
			
			Material* mat = 0;
			//mat = (Material*) (MaterialManager::getSingleton().getByName("GameTools/CubeMap/Default")).getPointer();
			
			if(i == 0) mat = (Material*) (MaterialManager::getSingleton().getByName("EnvMetals/Gold")).getPointer();
			else if(i == 1) mat = (Material*) (MaterialManager::getSingleton().getByName("EnvMetals/Silver")).getPointer();
			else if(i == 2) mat = (Material*) (MaterialManager::getSingleton().getByName("EnvMetals/Alu")).getPointer();
			else if(i == 3) mat = (Material*) (MaterialManager::getSingleton().getByName("EnvMetals/Copper")).getPointer();
			
			sphere->setMaterialName(mat->getName());
							
			OgreCubeMapRenderTechnique* techn = new OgreCubeMapRenderTechnique(0,
																				refreshinterval,
																				128,
																				0,
																				usecalculations,
																				usecalculations,
																				1.8,
																				5.5,
																				updateallfaces,
																				mat->getTechnique(0)->getPass(0),
																				rend,
																				group);
				
			OgreDistanceCubeMapRenderTechnique* techn2 = new OgreDistanceCubeMapRenderTechnique(0,
																				refreshinterval,
																				128,
																				1,
																				usecalculations,
																				usecalculations,
																				1.8,
																				5.5,
																				updateallfaces,
																				mat->getTechnique(0)->getPass(0),
																				rend,
																				group);	
			OgreDepthShadowRecieverRenderTechnique* shadowtechn = 
					new OgreDepthShadowRecieverRenderTechnique(1, 
													"GameTools/ShadowMap/ShadowVS",
													"GameTools/ShadowMap/ShadowPS",
													mat->getTechnique(0)->getPass(0),
													rend,
													group
													);
				
			group->addRenderTechnique(shadowtechn);
			group->addRenderTechnique(techn);
			group->addRenderTechnique(techn2);			
			sphere->getSubEntity(0)->setRenderTechniqueGroup(group);
		}

		{//athene
			OgreSharedRuns* sharedruns = new OgreSharedRuns();
			OgreIlluminationManager::getSingleton().addSharedRuns(sharedruns);		
			OgreTechniqueGroup* group = new OgreTechniqueGroup();
			group->addSharedRun(sharedruns);

			OgreRenderable* rend = new OgreRenderable(athene, 0);			
			sharedruns->addRenderable(rend);
			sharedruns->updateBounds();		
			
			//Material* mat = (Material*) athene->getSubEntity(0)->getMaterial().getPointer();
			
			
			Material* mat = (Material*) (MaterialManager::getSingleton().getByName("GameTools/CubeMap/Localized")).getPointer();
			
			athene->setMaterialName(mat->getName());

			

			OgreCausticCasterRenderTechnique* cautechn = 
				new OgreCausticCasterRenderTechnique(0,
													1,
													64,
													256,
													"GameTools/PhotonMapCaustic",
													"GameTools/Cau",
													0,
													false,
													true,
													mat->getTechnique(0)->getPass(0),
													rend,
													group
													);

			
			OgreCubeMapRenderTechnique* techn = new OgreCubeMapRenderTechnique(0,
																				refreshinterval,
																				128,
																				0,
																				true,
																				false,
																				0.8,
																				1.5,
																				updateallfaces,
																				mat->getTechnique(0)->getPass(0),
																				rend,
																				group);

			OgreDistanceCubeMapRenderTechnique* techn2 = new OgreDistanceCubeMapRenderTechnique(0,
																				refreshinterval,
																				128,
																				1,
																				true,
																				false,
																				0.8,
																				1.5,
																				updateallfaces,
																				mat->getTechnique(0)->getPass(0),
																				rend,
																				group);
			
			OgreDepthShadowRecieverRenderTechnique* shadowtechn = 
					new OgreDepthShadowRecieverRenderTechnique(1, 
													"GameTools/ShadowMap/ShadowVS",
													"GameTools/ShadowMap/ShadowPS",
													mat->getTechnique(0)->getPass(0),
													rend,
													group
													);
			
			group->addRenderTechnique(shadowtechn);				
			group->addRenderTechnique(techn);
			group->addRenderTechnique(techn2);
			group->addRenderTechnique(cautechn);
			athene->getSubEntity(0)->setRenderTechniqueGroup(group);
		}
		
   }

	void createFrameListener(void)
    {
		// This is where we instantiate our own frame listener
		mFrameListener= new RaytraceDemoListener(mWindow,
											mCamera,
											mainLightNode,
											goldSphereNode,
											silverSphereNode,
											copperSphereNode,
											aluSphereNode);
        mRoot->addFrameListener(mFrameListener);

    }

};

